"""Autobahn client implementation."""

from typing import Awaitable, Callable
import websockets
from autobahn_client.proto.message_pb2 import MessageType, PublishMessage, TopicMessage
import asyncio
from autobahn_client.util import Address


class Autobahn:
    def __init__(
        self,
        address: Address,
        reconnect: bool = True,
        reconnect_interval_seconds: float = 1.0,
    ):
        self.address = address
        self.websocket: websockets.ClientConnection | None = None
        self.first_subscription = True
        self.callbacks = {}
        self.reconnect = reconnect
        self.reconnect_interval_seconds = reconnect_interval_seconds
        self.listener_lock = asyncio.Lock()
        self.listener_task = None

    async def begin(self):
        try:
            self.websocket = await self.__connect()
        except OSError as e :
            print(f"Failed to connect to WebSocket at {self.address}: {str(e)}")
        
        if self.reconnect:
            asyncio.create_task(self.__maintain_connection())


    async def __connect(self) -> websockets.ClientConnection:    
        websocket = await websockets.connect(self.address.make_url())
        
        if self.callbacks and not self.first_subscription:
            self.__start_listener()

        return websocket

    async def __maintain_connection(self):
        while True:
            try:
                if self.websocket is None:
                    self.websocket = await self.__connect()
                    
                    for topic in self.callbacks.keys():
                        await self.websocket.send(
                            TopicMessage(
                                message_type=MessageType.SUBSCRIBE, topic=topic
                            ).SerializeToString()
                        )
                else:
                    try:
                        await self.websocket.ping()
                    except websockets.exceptions.ConnectionClosed:
                        self.websocket = None
                        print("Reconnecting...")
                await asyncio.sleep(self.reconnect_interval_seconds)
            except ConnectionError as e:
                print(f"Reconnection attempt failed: {str(e)}")
                await asyncio.sleep(self.reconnect_interval_seconds)

    async def ping(self):
        if self.websocket is None:
            raise ConnectionError("WebSocket not connected. Call begin() first.")

        await self.websocket.ping()

    async def publish(self, topic: str, payload: bytes):
        if self.websocket is None and not self.reconnect:
            raise ConnectionError("WebSocket not connected. Call begin() first.")
        
        if self.websocket is not None:
            message_proto = PublishMessage(
                message_type=MessageType.PUBLISH,
                topic=topic,
                payload=payload,
            )
            
            try:
                await self.websocket.send(message_proto.SerializeToString())
            except Exception as e:
                print(f"Error sending message: {str(e)}")
                self.websocket = None

    def __start_listener(self):
        if self.listener_task is None or self.listener_task.done():
            self.listener_task = asyncio.create_task(self.__listener())

    async def __listener(self):
        async with self.listener_lock:
            while True:
                try:
                    if self.websocket is None:
                        await asyncio.sleep(0.5)
                        continue

                    message = await self.websocket.recv()
                    if isinstance(message, str):
                        continue

                    message_proto = PublishMessage.FromString(message)
                    if message_proto.message_type == MessageType.PUBLISH:
                        if message_proto.topic in self.callbacks:
                            await self.callbacks[message_proto.topic](message_proto.payload)
                except websockets.exceptions.ConnectionClosed:
                    print("WebSocket connection closed, waiting for reconnection...")
                    self.websocket = None
                    await asyncio.sleep(0.5)
                    continue
                except Exception as e:
                    print(f"Error in listener: {str(e)}")
                    await asyncio.sleep(0.5)
                    continue

    async def subscribe(self, topic: str, callback: Callable[[bytes], Awaitable[None]]):
        if self.websocket is None and not self.reconnect:
            raise ConnectionError("WebSocket not connected. Call begin() first.")
        
        self.callbacks[topic] = callback
        
        if self.websocket is not None:
            await self.websocket.send(
                TopicMessage(
                    message_type=MessageType.SUBSCRIBE, topic=topic
                ).SerializeToString()
            )

        if self.first_subscription:
            self.__start_listener()
            self.first_subscription = False



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class MessageType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SERVER_STATE: _ClassVar[MessageType]
    SERVER_FORWARD: _ClassVar[MessageType]
    PUBLISH: _ClassVar[MessageType]
    SUBSCRIBE: _ClassVar[MessageType]
    UNSUBSCRIBE: _ClassVar[MessageType]
SERVER_STATE: MessageType
SERVER_FORWARD: MessageType
PUBLISH: MessageType
SUBSCRIBE: MessageType
UNSUBSCRIBE: MessageType

class AbstractMessage(_message.Message):
    __slots__ = ("message_type",)
    MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    message_type: MessageType
    def __init__(self, message_type: _Optional[_Union[MessageType, str]] = ...) -> None: ...

class ServerStateMessage(_message.Message):
    __slots__ = ("message_type", "uuid", "topics")
    MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    UUID_FIELD_NUMBER: _ClassVar[int]
    TOPICS_FIELD_NUMBER: _ClassVar[int]
    message_type: MessageType
    uuid: str
    topics: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, message_type: _Optional[_Union[MessageType, str]] = ..., uuid: _Optional[str] = ..., topics: _Optional[_Iterable[str]] = ...) -> None: ...

class ServerForwardMessage(_message.Message):
    __slots__ = ("message_type", "payload")
    MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    message_type: MessageType
    payload: bytes
    def __init__(self, message_type: _Optional[_Union[MessageType, str]] = ..., payload: _Optional[bytes] = ...) -> None: ...

class TopicMessage(_message.Message):
    __slots__ = ("message_type", "topic")
    MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    message_type: MessageType
    topic: str
    def __init__(self, message_type: _Optional[_Union[MessageType, str]] = ..., topic: _Optional[str] = ...) -> None: ...

class PublishMessage(_message.Message):
    __slots__ = ("message_type", "topic", "payload")
    MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    PAYLOAD_FIELD_NUMBER: _ClassVar[int]
    message_type: MessageType
    topic: str
    payload: bytes
    def __init__(self, message_type: _Optional[_Union[MessageType, str]] = ..., topic: _Optional[str] = ..., payload: _Optional[bytes] = ...) -> None: ...

class UnsubscribeMessage(_message.Message):
    __slots__ = ("message_type", "topic")
    MESSAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    TOPIC_FIELD_NUMBER: _ClassVar[int]
    message_type: MessageType
    topic: str
    def __init__(self, message_type: _Optional[_Union[MessageType, str]] = ..., topic: _Optional[str] = ...) -> None: ...



#!/usr/bin/env python3
"""Setup script for autobahn_client."""

import os
import subprocess
from setuptools import setup, find_packages, Command


class ProtocCommand(Command):
    """Custom command to run protoc compiler using make."""
    description = "run protoc compiler via make"
    user_options = []
    
    def initialize_options(self):
        pass
        
    def finalize_options(self):
        pass
    
    def run(self):
        command = ["make", "protoc"]
        self.announce("Running command: %s" % " ".join(command), level=2)
        subprocess.check_call(command)


if __name__ == "__main__":
    setup(
        packages=find_packages(),
        cmdclass={
            'protoc': ProtocCommand,
        },
    ) 


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: message.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'message.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rmessage.proto\x12\x0eproto.autobahn\"D\n\x0f\x41\x62stractMessage\x12\x31\n\x0cmessage_type\x18\x01 \x01(\x0e\x32\x1b.proto.autobahn.MessageType\"e\n\x12ServerStateMessage\x12\x31\n\x0cmessage_type\x18\x01 \x01(\x0e\x32\x1b.proto.autobahn.MessageType\x12\x0c\n\x04uuid\x18\x02 \x01(\t\x12\x0e\n\x06topics\x18\x03 \x03(\t\"Z\n\x14ServerForwardMessage\x12\x31\n\x0cmessage_type\x18\x01 \x01(\x0e\x32\x1b.proto.autobahn.MessageType\x12\x0f\n\x07payload\x18\x02 \x01(\x0c\"P\n\x0cTopicMessage\x12\x31\n\x0cmessage_type\x18\x01 \x01(\x0e\x32\x1b.proto.autobahn.MessageType\x12\r\n\x05topic\x18\x02 \x01(\t\"c\n\x0ePublishMessage\x12\x31\n\x0cmessage_type\x18\x01 \x01(\x0e\x32\x1b.proto.autobahn.MessageType\x12\r\n\x05topic\x18\x02 \x01(\t\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\"V\n\x12UnsubscribeMessage\x12\x31\n\x0cmessage_type\x18\x01 \x01(\x0e\x32\x1b.proto.autobahn.MessageType\x12\r\n\x05topic\x18\x02 \x01(\t*`\n\x0bMessageType\x12\x10\n\x0cSERVER_STATE\x10\x00\x12\x12\n\x0eSERVER_FORWARD\x10\x01\x12\x0b\n\x07PUBLISH\x10\x02\x12\r\n\tSUBSCRIBE\x10\x03\x12\x0f\n\x0bUNSUBSCRIBE\x10\x04\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'message_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_MESSAGETYPE']._serialized_start=569
  _globals['_MESSAGETYPE']._serialized_end=665
  _globals['_ABSTRACTMESSAGE']._serialized_start=33
  _globals['_ABSTRACTMESSAGE']._serialized_end=101
  _globals['_SERVERSTATEMESSAGE']._serialized_start=103
  _globals['_SERVERSTATEMESSAGE']._serialized_end=204
  _globals['_SERVERFORWARDMESSAGE']._serialized_start=206
  _globals['_SERVERFORWARDMESSAGE']._serialized_end=296
  _globals['_TOPICMESSAGE']._serialized_start=298
  _globals['_TOPICMESSAGE']._serialized_end=378
  _globals['_PUBLISHMESSAGE']._serialized_start=380
  _globals['_PUBLISHMESSAGE']._serialized_end=479
  _globals['_UNSUBSCRIBEMESSAGE']._serialized_start=481
  _globals['_UNSUBSCRIBEMESSAGE']._serialized_end=567
# @@protoc_insertion_point(module_scope)
